;; Copyright (c) 1996 John Zuckerman
;;
;; See the file "Notice" for information on usage and redistribution
;; of this file, and for a DISCLAIMER OF ALL WARRANTIES.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; Marks and Indexes
;; -----------------
;;
;; (see also scheme-text.ss)
;;
;; The fixed mark begin-exp is set to the beginning of the user input area.
;; begin-exp is set to the location of the end-exp mark immediately
;; upon a repl-port input operation.
;; 
;; The floating mark end-exp is set to the end of the user input area.
;; If user input is null, begin-exp equals end-exp.  
;;

;;
;; History System
;; --------------
;;
;; The user is always in one of two modes: browse mode or edit mode.
;; Browse mode occurs whenever the user is scrolling through the
;; history buffer.  Edit mode is the default, and is set whenever
;; the user has edited the current (unevaled) expression. The two
;; modes are mutually exclusive.
;;
;; On entering browse mode (Alt-P), if there is a current expression,
;; it is cached in the history buffer.  history-idx is set to the number
;; of the currently browsed item.
;;
;; On leaving browse mode for any reason, history-browse-mode? is set to #f
;; and any unevaled history item is discarded. The 0th history item
;; is the oldest.
;;

(require "../common/semaphore.ss" 'once)
(require "../common/scheme-text.ss" 'once)


(define-class (<repl-text> parent) (<scheme-text> parent)
  (ivars
   (repl-port-read-mode? #f)
   (rprm-begin-exp #f)
   (repl-port-read-extra-char #f)
   (chars-read 0)
   (write-mk #f)
   (my-thread #f)
   (exit-k #f)
   (history-list '())
   (history-size 0)
   (history-idx 0)
   (history-browse-mode? #f)
   (history-counter #f)
   (c/r-ready (thread-make-msg-queue 'c/r-ready))
   (waiter-reading-sem (let ((sem (make <sem> 1 'reading)))
                         (send sem wait)
                         sem))
   (ints-ok #t)
   (traffic-light #f)
   (repl-port #f) ;;; an input/output port 
   (ton-menu-item #f)
   (toff-menu-item #f)
   (copy-menu-item #f)
   (cut-menu-item #f)
   (paste-menu-item #f)
   (input-markup (create <markup> with
;                        (background-color: 'lightgoldenrodyellow)
;                        (background-color: 'white)
                         (foreground-color: 'blue)))
   )
  (inherited mini-buffer box-markup paren-markup highlight-markup
    key-prefix begin-exp end-exp end-mk)
  (inheritable mini-buffer box-markup paren-markup highlight-markup
    key-prefix begin-exp end-exp end-mk)

  (private
    [repl-port-read-mode (flag)
      (if flag
          (begin
            (set! rprm-begin-exp begin-exp)
            (set! repl-port-read-mode? #t))
          (set! repl-port-read-mode? #f))]
          
    [activate (mi)
      (send self turn-search-off)
      (when (send mi get-enabled) ((send mi get-action) mi))]
    [waiter-read ()
      (change-traffic 'green)
      (if (transcript-output-port)
          (begin 
            (send toff-menu-item set-enabled! #t)
            (send ton-menu-item set-enabled! #f))
          (begin
            (send ton-menu-item set-enabled! #t)
            (send toff-menu-item set-enabled! #f)))
      (send waiter-reading-sem signal)
      (thread-receive-msg c/r-ready)
      (mvlet ((dat sdat) (repl-read))
        (critical-section
          (change-traffic 'red)
          (send toff-menu-item set-enabled! #f)
          (send ton-menu-item set-enabled! #f)
          (send waiter-reading-sem wait)
          (when (or (zero? history-size)
                    (not (string=? (car history-list) sdat)))
            (history-add! sdat))
          dat))]

    [repl-read ()
      ;;
      ;; Returns two values:
      ;;
      ;;  * the next expression as a Scheme datum.
      ;;  * the next expression as a string obtained directly from the
      ;;    widget buffer.
      ;;
      (let* ([oldbe (send self fixed-mark begin-exp)]
             [old-cr chars-read]
             [dat (begin
                    (repl-port-read-mode #t)
                    (let ((val (read repl-port)))
                      (repl-port-read-mode #f)
                      val))]
             [slen (- chars-read old-cr)]
             [sdat (send self get-string oldbe
                         (send self add-offset oldbe slen))]
             )
        (let* ([newi (port-input-index repl-port)]
               [the-box (compute-the-box
                         oldbe
                         (send self add-offset oldbe slen)
                         input-markup)])
          ((car the-box)))
        (values dat sdat))]


    [compute-eos (str)
      ;;
      ;; returns the offset of the 1st character beyond the logical e.o.s.
      ;; (or zero if only whitespace).
      ;;
      (let loop ([j (fx1- (string-length str))])
        (cond [(fx= j -1) 0]
              [(char-whitespace? (string-ref str j)) (loop (fx1- j))]
              [else (fx1+ j)]))]

    [change-traffic (col)
      (when traffic-light (send traffic-light set-color! col))]

    [cursor () (send self get-cursor-pos)]

    [display-history-counter (count)
     (let ([pad
            (lambda (str len)
              (let ([slen (string-length str)])
                (if (< slen len)
                    (set! str
                      (string-append (make-string (- len slen) #\space) str)))
                (substring str 0 len)))])
     ; if history count exceeds 999, then
     ; this is more stable than we thought!
       (when history-counter
         (send history-counter set-title!
           (pad (or (and count (number->string count)) "") 3))))]

    [history-add! (str)
      (set! history-list (cons str history-list))
      (set! history-size (1+ history-size))]

    [history-delete! (idx)
      (set! history-list
        (let loop ((i (1- history-size)) (ls history-list))
          (if (null? ls)
            '()
            (if (= idx i)
              (loop (1- i) (cdr ls))
              (cons (car ls) (loop (1- i) (cdr ls)))))))
      (set! history-size (1- history-size))]

    [history-leave-browse-mode ()
;;; (fprintf (swl:bug-port) "history leave browse mode!!!~n")
      (when history-browse-mode?
        (let* ((bxi begin-exp)
               (beg (cons 0 (get-y/char bxi)))
               (end (idx1- bxi)))
          (send highlight-markup remove-markup self beg end))
        (history-delete! (1- history-size)) ;;; discard cached exp
        (display-history-counter #f)
        (history-terminate-search)
        (set! history-browse-mode? #f)
        )]

    [history-match (forward? anchored?)
;;;(fprintf (swl:bug-port) "history-match ~s ~s~n" forward? anchored?)
      (let ((sub-match
              (if anchored?
                (lambda (a1 a2 a3) (substring-match-anchored a1 a2 a3))
                (if forward?
                  (lambda (a1 a2 a3) (substring-match a1 a2 a3))
                  (lambda (a1 a2 a3) (substring-match-rev a1 a2 a3)))))
            (bump (if forward? 1+ 1-)))
        (define match
          (lambda (str start-idx start-ofs history-limit)
            ;;
            ;; returns pair (hist-idx . str-ofs) or #f if no match
            ;;
            (let loop ((h-idx start-idx) (ofs start-ofs))
              (if (or (negative? h-idx) (> h-idx history-limit))
                #f
                (let* ((hstr (list-ref history-list
                               (- (1- history-size) h-idx)))
                       (ofs (if ofs
                              (bump ofs)
                              (if (or anchored? forward?)
                                0
                                (1- (string-length hstr)))))
                       (match-ofs (sub-match str hstr ofs)))
                  (if match-ofs
                    (cons h-idx match-ofs)
                    (loop (bump h-idx) #f)))))))
        (unless search-string
          (set! search-string (send self get-string begin-exp end-exp))
          (set! search-offset #f))
        (let ((ofs (scan-until-newline search-string 0)))
          (send self display-mini
            (format "searching: ~a"
              (if ofs
                (string-append
                  (substring search-string 0 ofs) " ...")
                search-string))))
        (let* ((idx (if history-browse-mode?
                      history-idx
                      (if forward? 0 (1- history-size))))
               (limit (if history-browse-mode?
                        (- history-size 2)
                        (1- history-size)))
               (result (match search-string idx search-offset limit)))
          (if result
              (let ((ofs (cdr result)))
                (set! search-offset ofs)
                (history-move (cons 'abs (car result)))
                (if anchored? 
                    (send self cursor-eol)
                    (send self set-cursor-pos!
                          (add-offset self begin-exp ofs))))
              (send self display-mini "no matching history entry"))))]

    [history-move (n)
      ;;
      ;; +n/-n moves nth next/prev item into buffer
      ;; 'first/'last moves to the first/last item
      ;; (abs . n) moves the nth item into buffer
      ;;
      ;; also moves current exp into history, if necessary.
      ;;
;;;(fprintf (swl:bug-port) "history move n=~s hbm=~s~n"
;;;  n history-browse-mode?)
;;;(fprintf (swl:bug-port) "  history-size=~s~n" history-size)

      (unless history-browse-mode? ;;; not currently browsing
        (when (and (positive? history-size) ;;; there is history
                (or (and (number? n) (negative? n)) ;;; moving into it
                  (and (pair? n) (eq? (car n) 'abs))
                  (eq? n 'first)))
          (set! history-browse-mode? #t)
          (let* ((bxi begin-exp)
                 (beg (cons 0 (get-y/char bxi)))
                 (end (idx1- bxi)))
            (send highlight-markup apply-markup self beg end))
          (history-add!
            (send self get-string begin-exp end-exp)) ;;; cache cur exp
          (set! history-idx (1- history-size))))

      (when history-browse-mode?
        (let* ([new-pos
                 (cond
                   [(number? n) (+ history-idx n)]
                   [(and (pair? n) (eq? (car n) 'abs)) (cdr n)]
                   [(eq? n 'last) (1- history-size)]
                   [(eq? n 'first) 0]
                   [else (error 'history-move "invalid arg: ~s~n" n)])]
               [history-new (cond
                              [(negative? new-pos) 0]
                              [(< new-pos history-size) new-pos]
                              [else (1- history-size)])])
          (display-history-counter (1+ history-new))
          (unless (= history-idx history-new) ;;; already at end?
            (let* ((v-rows (send self get-height/char))
                   (v-top (yview-top-line))
                   (y-ins (get-y/char begin-exp))
                   (y-extra (- v-rows (- y-ins (1- v-top)))) ;;; extra lines
                   (str (list-ref history-list
                          (- (1- history-size) history-new))))
              (set! history-idx history-new)
              (let* ((str-lines (compute-lines str))
                     (pad-lines (+ (- y-extra str-lines) 2)) ;;; +2 empirical.
                     (pad-str (if (positive? pad-lines)
                                (make-string pad-lines #\newline)
                                ""))
                     (eos #f)
                     )
                (critical-section
                  (set! end-exp (send self floating-mark begin-exp))
                  ;;
                  ;; insert text w/o moving insert mark, else disturbs viewport.
                  ;;
                  (send self insert-at begin-exp str)
                  (set! eos (send self fixed-mark end-exp))
                  (send self insert-at eos pad-str)
                  (send self raw-delete end-exp end-mk)
                  (set! end-exp (send self floating-mark eos))
                  (send self set-cursor-pos! begin-exp)
                  (send self cursor-eol)
                  (show-paren-markup)
                  ))
              (when (= history-idx (1- history-size))
                (history-leave-browse-mode))
              ))))]

    [open-repl-port ()
      (let* ([ibuf (make-string 100)]
             [obuf (make-string 2000)]
             [ibuf-s (string-length ibuf)]
             [ibuf-i 0]
             [ibuf-n 0]
             [obuf-s (string-length obuf)]
             [obuf-i 0] ;;; pseudo output-index 
             [line-buf (open-input-string "")]
             [flush-output-thread #f]
             [seen-newline-sem (make <sem> 1 'sem:newline)]
             [name "repl"]
             [eof-msg "unexpected end-of-file"]
             [emlen (string-length eof-msg)]
             )

        (define well-formed-exp?
          (lambda (default-error-handler)
            (call/cc
             (lambda (escape)
               (define read-error-handler
                 (lambda (who msg . args)
                   (if (and (pair? args)
                            (string? (car args))
                            (>= (string-length (car args)) emlen)
                            (string=? (substring (car args) 0 emlen)
                                      eof-msg))
                       (escape #f) ;;; user may still be typing
                       (escape #t) ;;; allow read of repl-port to handle error
                       )))
               (let ([p (open-input-string
                         (send self get-string rprm-begin-exp end-exp))])
                 (parameterize ([error-handler read-error-handler])
                   (not (eof-object? (read p)))))))))
        ;;
        ;; don't use trace-lambda unless trace output port properly bound
        ;;
        (define widget-block-read
          ;;
          ;; Reads lines one at a time from buffer.
          ;; Returns num chars placed in str.  May block on c/r-ready.
          ;;
          (lambda (str cnt)
            (unless repl-port
              (error 'widget-block-read "port already closed"))
            (let ((n (block-read line-buf str cnt)))
              (if (not (eof-object? n))
                  (begin 
                    ;; this will only ensure we copy into the transcript
                    ;; what user types after evaluating something like
                    ;; (read) while in the repl.
                    (when (and (transcript-output-port) (fixnum? n))
                      (block-write (transcript-output-port) str n))
                    n)
                  (let loop ()
                    (critical-section
                      ;; Get input one line at a time, so that multiple
                      ;; expression key-ahead works.
                      (let* ([bxpos (send begin-exp get-pos)]
                             [expos (send end-exp get-pos)]
                             [bxl (cdr bxpos)]
                             [exl (cdr expos)]
                             [lnpos (if (= bxl exl) expos `(0 . ,(1+ bxl)))])
;;;(fprintf (swl:bug-port) "bxpos=~s expos=~s bxl=~s exl=~s lnpos=~s~n"
;;;      bxpos expos bxl exl lnpos)
                        (if (or (equal? bxpos expos)
                                (and repl-port-read-mode?
                                     (not (well-formed-exp? (error-handler)))))
                            (begin
                              (thread-receive-msg c/r-ready)
                              (loop))
                            (begin
                              (set! begin-exp (send self fixed-mark lnpos))
                              (set! write-mk (send self fixed-mark lnpos))
                              (let ((ln (send self get-string bxpos lnpos)))
;;;(fprintf (swl:bug-port) "ln=~s~n" ln)                              
                                (set! line-buf (open-input-string ln)))
                              (widget-block-read str cnt))))))))))

        (define widget-block-write
          (lambda (buf i)
            (critical-section
              (when repl-port
                (set! write-mk (send self floating-mark write-mk))
                (set! begin-exp (send self floating-mark begin-exp))
                (if (eqv? (string-length buf) i)
                    (send self unrestricted-insert-at write-mk buf)
                    (send self unrestricted-insert-at write-mk
                          (substring buf 0 i)))
                (when (transcript-output-port)
                  (block-write (transcript-output-port) buf i))
                (set! write-mk (send self fixed-mark write-mk))
                (set! begin-exp (send self fixed-mark begin-exp))
                (send self make-visible end-exp)
                ))))
        
        (define console-read-flush ;;; mimics chez console port behavior
          (lambda (p)
            (when (eq? (console-input-port) p)
              (flush-output-port (console-output-port))
              )))
        
        (define handler
          (lambda (xmsg . xargs)
              (record-case (cons xmsg xargs)
                [block-write (p str i)
                 (flush-output-port p)
                 (widget-block-write str i)]
                [write-char (c p)
                  (disable-interrupts) ;;; want max efficiency here
                    (string-set! obuf obuf-i c) (set! obuf-i (fx1+ obuf-i))
                  (enable-interrupts)
                  (if (= obuf-i obuf-s)
                      (flush-output-port p)
                      (when (char=? c #\newline)
                        (send seen-newline-sem signal)
                        (void)))]
                [file-position (p . pos)
                 (if (null? pos)
                     chars-read
                     (error 'repl-port "cannot reposition"))]
                [flush-output-port (p)
                   (critical-section
                     (when (fx> obuf-i 0)
                       (widget-block-write obuf obuf-i)
                       (set! obuf-i 0)))]
                [close-port (p)
                  (thread-kill flush-output-thread)
                  (set-port-output-size! p 0)
                   (set-port-input-size! p 0)
                   (mark-port-closed! p)]
                [port-name (p) name]
                [read-char (p)
                  (critical-section
                    (let ([c (peek-char p)])
                      (set! chars-read (fx1+ chars-read))
                      (set! ibuf-i (fx1+ ibuf-i))
                      c))]
                [peek-char (p)
                  (critical-section
                    (console-read-flush p)
                    (let ((ans
                           (if (fx< ibuf-i ibuf-n)
                               (string-ref ibuf ibuf-i)
                               (let ([n (widget-block-read ibuf ibuf-s)])
                                 (set! ibuf-i 0)
                                 (set! ibuf-n n)
                                 (string-ref ibuf 0)))))
                      ans))]
                [unread-char (c p)
                  (critical-section
                    (when (fx= ibuf-i 0)
                      (error 'unread-char "tried to unread too far on ~s" p))
                    (set! ibuf-i (fx1- ibuf-i))
                    (set! chars-read (fx1- chars-read)))]
                [char-ready? (p)
                  (let ((ans (or (fx< ibuf-i ibuf-n)
                                 (not (send self pos=? begin-exp end-exp)))))
                    ans)]
                [clear-input-port (p)
                  (critical-section
                    (set! ibuf-i 0)
                    (set! ibuf-n 0)
                    (set! line-buf (open-input-string "")))]
                [clear-output-port (p) (set-port-output-index! p 0)]
                [else (error 'repl-port "operation ~s not handled" xmsg)])))
        
        (let ([iop (make-input/output-port handler ibuf obuf)])
          (set-port-input-size! iop 0) ;;; always invoke handler
          (set-port-output-size! iop 0) ;;; always invoke handler
          (set! flush-output-thread
            (thread-fork
              (lambda ()
                (let f ()
                  ;;
                  ;; This is not the fairness hack - it slows down the
                  ;; flush rate, improving performance.
                  ;;
                  (thread-name "flush")
                  (thread-sleep 100)
                  (send seen-newline-sem wait)
                  (flush-output-port iop)
                  (f)))))
          (set! repl-port-read-extra-char
            (lambda ()
              (if (fx< ibuf-i ibuf-n)
                  (let ((ch (string-ref ibuf ibuf-i)))
                    (set! ibuf-i (fx1+ ibuf-i))
                    ch)
                  #f)))
          iop))]

    [history-terminate-search ()
      (set! search-string #f)
      (send self clear-mini)]

    [yview-top-line ()
      (1+ ;;; base is relative to 1
        (inexact->exact
          (round
            (* (get-y/char end-mk) (car (send self get-vertical-view))))))]

    [port-alive? () (and repl-port (not (port-closed? repl-port)))]

    [remove-unprintables (str)
     ;;
     ;; remove things that upset the Tk text widget.
     ;;                  
     (list->string                       
      (let f ((ls (string->list str)))
        (cond [(null? ls) '()]
              [(< (char->integer (car ls)) (char->integer #\newline))
               (f (cdr ls))]
              [else (cons (car ls) (f (cdr ls)))])))]

    ) ;;; end private

  (protected ;;;        ** ** PROTECTED ** **

    )
  
  (public ;;;            ** ** PUBLIC ** **

    [init (parent)
      (send-base self init parent)
      (set! write-mk (send self fixed-mark '(0 . 0)))
    ]

    [set-menu-items! (mi-toff mi-ton mi-copy mi-cut mi-paste)
      (set! toff-menu-item mi-toff)
      (set! ton-menu-item mi-ton)
      (set! copy-menu-item mi-copy)
      (set! cut-menu-item mi-cut)
      (set! paste-menu-item mi-paste)
      ]

    [transcript-on (tname)
      (when (transcript-output-port)
        (error 'repl "transcript port open"))
      (send self set-current-expression!
        (string-append "(transcript-on \"" tname "\")
"))
      (thread-send-msg c/r-ready #t)]

    [transcript-off ()
      (unless (transcript-output-port)
        (error 'repl "transcript port not open"))
      (send self set-current-expression!
        (string-append "(transcript-off)
"))
      (thread-send-msg c/r-ready #t)]

    [destroy ()
      (close-port repl-port)
      (unless (eq? thread-r my-thread) ;;; may be called by my-thread
        (thread-break my-thread 'shutdown exit-k)
        (thread-yield))
      (console-error-port (swl:error-log-port))
      (send-base self destroy)
      ]

    [set-current-expression! (str)
      (critical-section
        (send waiter-reading-sem wait)
        (send self delete begin-exp end-exp)
        (send self insert str)
        (send waiter-reading-sem signal)
        (send self set-cursor-pos! end-exp)
        )]

    [history-first () (history-move 'first)]
    [history-last () (history-move 'last)]
    [history-backward () (history-move -1)]
    [history-forward () (history-move +1)]
    
    [history-get ()
      ;; the car of history-list is the last history item.
      history-list]

    [history-set! (hi)
      (set! history-list hi)
      (set! history-size (length hi))
      (set! history-idx 0)
      ]

    [history-clear ()
      (history-leave-browse-mode)
      (set! history-list '())
      (set! history-size 0)
      (set! history-idx 0)]

    [key-press-prefix (key mods)
      (define hist-mat
        (lambda (f1 f2)
          (send self clear-mini)
          (history-match f1 f2)))
      
      (case key-prefix
        [(escape)
         (set! key-prefix #f)
         (event-case ((key= key) (modifier= mods))
           [([#\P]) (hist-mat #f #f)]
           [([#\p]) (hist-mat #f #t)]
           [([#\N]) (hist-mat #t #f)]
           [([#\n]) (hist-mat #t #t)]
           [([shift_l] [shift_r] [control_l] [control_r])
            (set! key-prefix 'escape) ;;; doesn't undo an escape
            ]
           [else
             (send self display-mini "unknown escape command")])]
        [else
          (send-base self key-press-prefix key mods)])
      ]

    [key-press-no-prefix (key mods)
      (event-case ((key= key) (modifier= mods))
        [([alt])
         (event-case ((key= key) (modifier= mods))
           [([alt #\q]) (send (send self get-parent) destroy)]
           [([alt #\n]) (send self history-forward)]
           [([alt #\p]) (send self history-backward)]
           [([alt #\c]) (activate copy-menu-item)]
           [([alt #\x]) (activate cut-menu-item)]
           [([alt #\v]) (activate paste-menu-item)]
           [else
            (send-base self key-press-no-prefix key mods)])]
        [([control] [up] [down] [#\backspace] [delete])
         (event-case ((key= key) (modifier= mods))
           [([control #\c]) (send self interrupt)]
           [([control #\d] [delete] [control #\k])
            (history-leave-browse-mode)
            (history-terminate-search)
            (send-base self key-press-no-prefix key mods)]
           [([control #\h] [#\backspace])
            (when (send self pos>? (cursor) begin-exp)
              (history-leave-browse-mode)
              (history-terminate-search)
              (send-base self key-press-no-prefix key mods))]
           [([control #\n] [down])
            (send self turn-search-off)
            (let* ([end-line (get-y/char (idx1- end-exp))]
                   [cur-line (get-y/char (cursor))])
              (if (and (= cur-line end-line)
                       (or history-browse-mode?
                           (send self pos=? begin-exp end-exp)))
                (send self history-forward)
                (send-base self key-press-no-prefix key mods)
                  ))]
           [([control #\p] [up])
            (send self turn-search-off)
            (let* ([begin-line (get-y/char begin-exp)]
                   [cur-line (get-y/char (cursor))])
              (if (and (= cur-line begin-line)
                       (or history-browse-mode?
                           (send self pos=? begin-exp end-exp)))
                (send self history-backward)
                (send-base self key-press-no-prefix key mods)
                  ))]
           [else
            (send-base self key-press-no-prefix key mods)])]
        [([escape])
         (set! key-prefix 'escape)
         (send self display-mini "Key prefix: Esc")
         ]
;       [([#\return] [#\newline])
;        (send-base self key-press-no-prefix key mods)
;        (when (str-whitespace? (send self get-string (cursor) end-exp))
;           (thread-send-msg c/r-ready #t))
;        ]
        [else
         (send-base self key-press-no-prefix key mods)])
      ]

    [interrupt ()
      (send self turn-search-off)
      (when (port-alive?)
        (critical-section
          (history-leave-browse-mode)
          (send self raw-delete end-exp end-mk)
          (set! begin-exp (send self fixed-mark end-exp))
          (set! write-mk (send self fixed-mark end-exp))
          (send self set-cursor-pos! end-exp)
          (thread-break my-thread 'interrupt)))]

    [delete-eol ()
      (let ((ofs (cursor)))
        (when (and (send self pos<=? begin-exp ofs)
                   (null? (send self markups-at ofs)))
          (send-base self delete-eol)))]

    [delete-char (disp)
      (let ((ofs (send self add-offset (cursor) disp)))
        (when (and (send self pos<=? begin-exp ofs end-exp)
                   (null? (send self markups-at ofs)))
          (send-base self delete-char disp)))]

    [delete (idx)
      (when (and (send self pos<=? begin-exp idx end-exp)
                 (null? (send self markups-at (cursor))))
        (send-base self delete idx))]

    [delete (beg end)
      (when (and (send self pos<=? begin-exp beg)
                 (send self pos<=? end end-exp)
                 (null? (send self markups-at (cursor))))
        (send-base self delete beg end))]

    [raw-delete (beg end)
      (send-base self delete beg end)]

    [insert (str)
     ;;
     ;; avoids markup bug that causes markup to bleed between ()
     ;;
     (if (and (send self pos<=? begin-exp (cursor) end-exp)
              (null? (send self markups-at (cursor))))
         (begin
           (remove-paren-markup)
           (history-leave-browse-mode)
           (history-terminate-search)
           (send-base self insert str)
           (show-paren-markup)
           ;; if user pastes a selection that ends in a newline (+ whitespace)
           ;; process as if the newline had been typed by user.
           (when (and (or (memv str '(#\newline #\return))
                          (and (string? str)
                               (not (string=? str ""))
                               (let nl? ([i (fx- (string-length str) 1)])
                                 (and (fx>= i 0)
                                      (let ([c (string-ref str i)])
                                        (or (char=? c #\newline)
                                            (and (char-whitespace? c)
                                                 (nl? (fx- i 1)))))))))
                      (str-whitespace? (send self get-string (cursor) end-exp)))
             (thread-send-msg c/r-ready #t)))
         (begin
;          (send self display-mini
;            (string-append "Input not allowed here."))
           (send self set-cursor-pos! end-exp)))]
;;;             (let ((o (open-output-string)))
;;;               (display "cursor: " o)(display (cursor) o)
;;;               (display "begin-exp: " o) (send begin-exp print o)
;;;               (display "end-exp: " o) (send end-exp print o)
;;;               (get-output-string o)))))]

    [repl-new-cafe (hcount)
      (set! my-thread (thread-self))
      (set! repl-port (open-repl-port))
      (set! history-counter hcount)
      (display-history-counter #f) ; after set! of history-counter
      (send self set-wrap! 'none)
      (send self set-focus)
      ;;
      ;; do not use parameterize here, or user-forked threads won't
      ;; pick up the correct values.
      ;; 
      (console-input-port repl-port) (console-output-port repl-port)
      (console-error-port repl-port)
      (current-input-port repl-port) (current-output-port repl-port)
      (when (top-level-bound? 'trace-output) (trace-output repl-port))
      (when (top-level-bound? 'trace-output-port)
        (trace-output-port repl-port))

      (fprintf (console-output-port)
       "~aChez Scheme Version ~a~%Copyright (c) 1998 Cadence Research Systems~%~%"
       (if (system-ref '\#compiler-is-loaded?) "" "Petite ")
       (system-ref '\#scheme-version))

      (waiter-prompt-and-read
       (lambda (n)
         (critical-section
           (flush-output-port repl-port) ;;; needed for correct oldwmk
           (let* ([r
                   (list->string
                    (reverse
                     (let f ((acc '()))
                       (let ((ch (repl-port-read-extra-char)))
                         (cond
                           [(not (char? ch)) acc]
                           [(char=? ch #\newline)
                            ;; if msg waiting, then the remainder is already
                            ;; available to the reader, hence newline belongs
                            ;; in the remainder.
                            (if (thread-msg-waiting? c/r-ready)
                                (cons #\newline acc)
                                acc)]
                           [else (f (cons ch acc))])))))]
                  [remainder (if (= (compute-eos r) 0) "" r)]
                  [oldwmk (send self fixed-mark write-mk)])
;;;(fprintf (swl:bug-port) "rem=~s r=~s~n" remainder r)
             (let loop ([n n])
               (unless (fxzero? n)
                 (display (waiter-prompt-string))
                 (loop (fx- n 1))))
             (display #\space)
             (flush-output-port repl-port)
             (set! write-mk oldwmk)
             (send self insert-at begin-exp remainder)
             (send self make-visible end-exp)
             ))
         (let ([x (waiter-read)])
           (unless (eof-object? x) (send self raw-delete end-exp end-mk))
           x)))

      (waiter-write
       (lambda (x)
         (unless (eq? x (void))
           (when (string? x) (set! x (remove-unprintables x)))
           (pretty-print x (console-output-port))
         (flush-output-port (console-output-port)))))
      
      (call/cc
       (lambda (k)
         (set! exit-k k)
         (new-cafe)))
      (set! exit-k #f) ;;; see destroy method
      (void)]

    [set-traffic-light! (tl) (set! traffic-light tl)]

        ) ;;; end public
      )

#!eof
